/*****************************************************************************
 *
 * FILE:	parsemail.c
 * DESCRIPTION:	parse mail
 * DATE:	Thu, Jun  8 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2001 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: parsemail.c,v 1.7 2001/06/29 10:08:10 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include "mailtools.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
typedef enum {
  MHF_START,	/* start of header */
  MHF_DATE,
  MHF_FROM,
  MHF_SENDER,
  MHF_REPLY_TO,
  MHF_TO,
  MHF_CC,
  MHF_BCC,
  MHF_MESSAGE_ID,
  MHF_IN_REPLY_TO,
  MHF_REFERENCES,
  MHF_SUBJECT,
  MHF_COMMENTS,
  MHF_KEYWORDS,
  MHF_RETURN_PATH,
  MHF_RECEIVED,
  MHF_MIME_VERSION,
  MHF_CONTENT_TYPE,
  MHF_CONTENT_TRANSFER_ENCODING,
  MHF_CONTENT_ID,
  MHF_CONTENT_DESCRIPTION,
  MHF_CONTENT_DISPOSITION,
  MHF_OTHERS,
  MHF_END	/* end of header */
} HeaderField;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void			parseMailHeader(Mail *);
static void			parseMailBody(Mail *);
static String			getFileName(Mail *);
static MultiPart *		newMultiPart(MultiPart *, Mail *);
static Received *		newReceivedFiled(String, Received *);
static ContentType *		getContentType(String);
static ContentDisposition *	getContentDisposition(String);
static String			commaSplit(String);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	mailOpen
 * Description:		open the mail file
 * Arguments:		mail filename, option
 * Return Value:	Mail pointer
 */
Mail *
mailOpen(filename, option)
	String		filename;
	option_t	option;
{
  Mail *	mail	= NULL;
  String	p	= NULL;

  mail = (Mail *)calloc(1, sizeof(Mail));
  if (mail == NULL) {
    return NULL;
  }

  mail->header = (MailHeader *)calloc(1, sizeof(MailHeader));
  if (mail->header == NULL) {
    free((Mail *)mail); mail = NULL;
    return NULL;
  }

  mail->body = (MailBody *)calloc(1, sizeof(MailBody));
  if (mail->body == NULL) {
    free((MailHeader *)mail->header); mail->header = NULL;
    free((Mail *)mail); mail = NULL;
    return NULL;
  }

  /* open the mail file */
  if (option == OPT_FILTER) {
    mail->fin = stdin;
  }
  else {
    mail->fin = fopen(filename, "r");
    if (mail->fin == NULL) {
      free((MailBody *)mail->body); mail->body = NULL;
      free((MailHeader *)mail->header); mail->header = NULL;
      free((Mail *)mail); mail = NULL;
      return NULL;
    }
  }

  /* extract the base portion of filename */
  p = strrchr(filename, '/');
  if (p != NULL) {
    p++;
    mail->filename = strdup(p);
  }
  else {
    mail->filename = strdup(filename);
  }

  mail->parts	  = 0;
  mail->option	  = option;
  mail->multipart = NULL;

  parseMailHeader(mail);
  parseMailBody(mail);

  return mail;
}

/*
 * Function Name:	mailClose
 * Description:		close the mail file and release allocated memory area
 * Arguments:		Mail pointer
 * Return Value:	none
 */
void
mailClose(mail)
	register Mail *	mail;
{
#define	FREE_STRING(x)	if((x)){free((String)(x));(x)=NULL;}
#define	CLOSE_FILE(x)	if((x)){fclose((x));(x)=NULL;}
  register MailHeader *	header	= NULL;
  register Received *	rp	= NULL;
  register Received *	rp_tmp	= NULL;
  register MailBody *	body	= NULL;
  register MultiPart *	mp	= NULL;
  register MultiPart *	mp_tmp	= NULL;

  if (mail == NULL) {
    return;
  }
  FREE_STRING(mail->filename);
  CLOSE_FILE(mail->fin);

  /* for header */
  header = mail->header;
  FREE_STRING(header->date);
  FREE_STRING(header->from);
  FREE_STRING(header->sender);
  FREE_STRING(header->reply_to);
  FREE_STRING(header->to);
  FREE_STRING(header->cc);
  FREE_STRING(header->bcc);
  FREE_STRING(header->message_id);
  FREE_STRING(header->in_reply_to);
  FREE_STRING(header->references);
  FREE_STRING(header->subject);
  FREE_STRING(header->comments);
  FREE_STRING(header->keywords);
  FREE_STRING(header->return_path);
  FREE_STRING(header->mime_version);
  FREE_STRING(header->content_transfer_encoding);
  FREE_STRING(header->content_id);
  FREE_STRING(header->content_description);
  for (rp = header->received; rp != NULL; rp = rp_tmp) {
    rp_tmp = rp->next;
    FREE_STRING(rp->recv_from);
    FREE_STRING(rp->recv_by);
    FREE_STRING(rp->recv_via);
    FREE_STRING(rp->recv_with);
    FREE_STRING(rp->recv_id);
    FREE_STRING(rp->recv_for);
    FREE_STRING(rp->recv_date);
    free((Received *)rp);
  }
  if (header->content_type) {
    FREE_STRING(header->content_type->media_type);
    FREE_STRING(header->content_type->charset);
    FREE_STRING(header->content_type->boundary);
    FREE_STRING(header->content_type->name);
    FREE_STRING(header->content_type->others);
  }

  /* for body */
  body = mail->body;
  FREE_STRING(body->filename);
  CLOSE_FILE(body->fout);
  if (body->tmpfile) {
    (void)unlink(body->tmpfile);
    free((String)body->tmpfile); body->tmpfile = NULL;
  }
  CLOSE_FILE(body->ftmp);

  /* for multipart */
  for (mp = mail->multipart; mp != NULL; mp = mp_tmp) {
    mp_tmp = mp->next;
    FREE_STRING(mp->filename);
    CLOSE_FILE(mp->fout);
    if (mp->tmpfile) {
      (void)unlink(mp->tmpfile);
      free((String)mp->tmpfile); mp->tmpfile = NULL;
    }
    CLOSE_FILE(mp->ftmp);
    if (mp->content_type) {
      FREE_STRING(mp->content_type->media_type);
      FREE_STRING(mp->content_type->charset);
      FREE_STRING(mp->content_type->boundary);
      FREE_STRING(mp->content_type->name);
      FREE_STRING(mp->content_type->others);
    }
    FREE_STRING(mp->content_transfer_encoding);
    if (mp->content_disposition) {
      FREE_STRING(mp->content_disposition->disposition);
      FREE_STRING(mp->content_disposition->filename);
      FREE_STRING(mp->content_disposition->creation_date);
      FREE_STRING(mp->content_disposition->modification_date);
      FREE_STRING(mp->content_disposition->read_date);
      FREE_STRING(mp->content_disposition->others);
    }
    FREE_STRING(mp->content_description);
    free((MultiPart *)mp);
  }
}

/*****************************************************************************
 *****************************************************************************/


/*
 * Function Name:	parseMailHeader
 * Description:		parse mail header
 * Arguments:		Mail pointer
 * Return Value:	none
 */
static void
parseMailHeader(mail)
	Mail *	mail;
{
  MailHeader *	header		= mail->header;
  String	value		= NULL;
  String	content_type	= NULL;	/* for Content-Type field */
  String	received	= NULL;	/* for Received field */
  char		buf[BUFSIZ];
  HeaderField	field;

  field = MHF_START;
  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), mail->fin) != NULL) {
    if (buf[0] == LF || (buf[0] == CR && buf[1] == LF)) {
      header->content_type = getContentType(content_type);
      break;	/* end of header */
    }

    /* continued value */
    if (buf[0] == SPC || buf[0] == TAB) {
      for (value = buf; *value == SPC || *value == TAB; value++) {
	;	/* remove white spaces */
      }
      switch (field) {
	case MHF_DATE:
	  header->date = append(chomp(value), header->date);
	  continue;
	case MHF_FROM:
	  header->from = append(chomp(value), header->from);
	  continue;
	case MHF_SENDER:
	  header->sender = append(chomp(value), header->sender);
	  continue;
	case MHF_REPLY_TO:
	  header->reply_to = append(chomp(value), header->reply_to);
	  continue;
	case MHF_TO:
	  header->to = append(chomp(value), header->to);
	  continue;
	case MHF_CC:
	  header->cc = append(chomp(value), header->cc);
	  continue;
	case MHF_BCC:
	  header->bcc = append(chomp(value), header->bcc);
	  continue;
	case MHF_MESSAGE_ID:
	  header->message_id = append(chomp(value), header->message_id);
	  continue;
	case MHF_IN_REPLY_TO:
	  header->in_reply_to = append(chomp(value), header->in_reply_to);
	  continue;
	case MHF_REFERENCES:
	  header->references = append(chomp(value), header->references);
	  continue;
	case MHF_SUBJECT:
	  header->subject = append(chomp(value), header->subject);
	  continue;
	case MHF_COMMENTS:
	  header->comments = append(chomp(value), header->comments);
	  continue;
	case MHF_KEYWORDS:
	  header->keywords = append(chomp(value), header->keywords);
	  continue;
	case MHF_RETURN_PATH:
	  header->return_path = append(chomp(value), header->return_path);
	  continue;
	case MHF_RECEIVED:
	  received = append(chomp(value), received);
	  continue;
	case MHF_MIME_VERSION:
	  header->mime_version = append(chomp(value), header->mime_version);
	  continue;
	case MHF_CONTENT_TYPE:
	  content_type = append(chomp(value), content_type);
	  continue;
	case MHF_CONTENT_TRANSFER_ENCODING:
	  header->content_transfer_encoding =
		append(chomp(value), header->content_transfer_encoding);
	  continue;
	case MHF_CONTENT_ID:
	  header->content_id = append(chomp(value), header->content_id);
	  continue;
	case MHF_CONTENT_DESCRIPTION:
	  header->content_description =
		append(chomp(value), header->content_description);
	  continue;
	default:
      }
    }

    if (received != NULL) {
      header->received = newReceivedFiled(received, header->received);
      received = NULL;
    }

    value = strchr(buf, ':');
    if (value != NULL) {
      for (++value; *value == SPC || *value == TAB; value++) {
	;	/* remove white spaces */
      }
    }

    chomp(value);
    if (strncasecmp(buf, MH_DATE, length(MH_DATE)) == 0) {
      header->date	= strdup(value);
      field		= MHF_DATE;
    }
    else if (strncasecmp(buf, MH_FROM, length(MH_FROM)) == 0) {
      header->from	= strdup(value);
      field		= MHF_FROM;
    }
    else if (strncasecmp(buf, MH_SENDER, length(MH_SENDER)) == 0) {
      header->sender	= strdup(value);
      field		= MHF_SENDER;
    }
    else if (strncasecmp(buf, MH_REPLY_TO, length(MH_REPLY_TO)) == 0) {
      header->reply_to	= strdup(value);
      field		= MHF_REPLY_TO;
    }
    else if (strncasecmp(buf, MH_TO, length(MH_TO)) == 0) {
      header->to	= strdup(value);
      field		= MHF_TO;
    }
    else if (strncasecmp(buf, MH_CC, length(MH_CC)) == 0) {
      header->cc	= strdup(value);
      field		= MHF_CC;
    }
    else if (strncasecmp(buf, MH_BCC, length(MH_BCC)) == 0) {
      header->bcc	= strdup(value);
      field		= MHF_BCC;
    }
    else if (strncasecmp(buf, MH_MESSAGE_ID, length(MH_MESSAGE_ID)) == 0) {
      header->message_id	= strdup(value);
      field			= MHF_MESSAGE_ID;
    }
    else if (strncasecmp(buf, MH_IN_REPLY_TO, length(MH_IN_REPLY_TO)) == 0) {
      header->in_reply_to	= strdup(value);
      field			= MHF_IN_REPLY_TO;
    }
    else if (strncasecmp(buf, MH_REFERENCES, length(MH_REFERENCES)) == 0) {
      header->references	= strdup(value);
      field			= MHF_REFERENCES;
    }
    else if (strncasecmp(buf, MH_SUBJECT, length(MH_SUBJECT)) == 0) {
      header->subject	= strdup(value);
      field		= MHF_SUBJECT;
    }
    else if (strncasecmp(buf, MH_COMMENTS, length(MH_COMMENTS)) == 0) {
      header->comments	= strdup(value);
      field		= MHF_COMMENTS;
    }
    else if (strncasecmp(buf, MH_KEYWORDS, length(MH_KEYWORDS)) == 0) {
      header->keywords	= strdup(value);
      field		= MHF_KEYWORDS;
    }
    else if (strncasecmp(buf, MH_RETURN_PATH, length(MH_RETURN_PATH)) == 0) {
      header->return_path	= strdup(value);
      field			= MHF_RETURN_PATH;
    }
    else if (strncasecmp(buf, MH_RECEIVED, length(MH_RECEIVED)) == 0) {
      received	= strdup(value);
      field	= MHF_RECEIVED;
    }
    else if (strncasecmp(buf, MH_MIME_VERSION, length(MH_MIME_VERSION)) == 0) {
      header->mime_version	= strdup(value);
      field			= MHF_MIME_VERSION;
    }
    else if (strncasecmp(buf, MH_CONTENT_TYPE,
			 length(MH_CONTENT_TYPE)) == 0) {
      content_type	= strdup(value);
      field		= MHF_CONTENT_TYPE;
    }
    else if (strncasecmp(buf, MH_CONTENT_TRANSFER_ENCODING,
			 length(MH_CONTENT_TRANSFER_ENCODING)) == 0) {
      header->content_transfer_encoding	= strdup(value);
      field				= MHF_CONTENT_TRANSFER_ENCODING;
    }
    else if (strncasecmp(buf, MH_CONTENT_ID, length(MH_CONTENT_ID)) == 0) {
      header->content_id	= strdup(value);
      field			= MHF_CONTENT_ID;
    }
    else if (strncasecmp(buf, MH_CONTENT_DESCRIPTION,
			 length(MH_CONTENT_DESCRIPTION)) == 0) {
      header->content_description	= strdup(value);
      field				= MHF_CONTENT_DESCRIPTION;
    }
    else {
      field = MHF_OTHERS;
    }
  }	/* end of while() */
}

/*
 * Function Name:	parseMailBody
 * Description:		parse mail body
 * Arguments:		Mail pointer
 * Return Value:	none
 */
static void
parseMailBody(mail)
	Mail *	mail;
{
  MailBody *	body	= mail->body;
  MailHeader *	header	= mail->header;
  char		buf[BUFSIZ];

  /* set filename for mail body */
  body->filename = getFileName(mail);
  body->tmpfile	 = strdup(TEMPLATE_FILE);
  if (mail->option != OPT_PRINT && body->filename && body->tmpfile) {
    int	fd = mkstemp(body->tmpfile);

    if (fd == -1) {
      error();
    }
    body->ftmp = fdopen(fd, "w+");
    if (body->ftmp == NULL) {
      error();
    }
    (void)unlink(body->tmpfile);
  }

  /* parse multipart body */
  if (header->content_type && header->content_type->boundary) {
    MultiPart *	multipart		= mail->multipart;
    String	boundary		= NULL;
    String	bend			= NULL;	/* end of multipart body */
    String	value			= NULL;
    String	content_type		= NULL;
    String	content_disposition	= NULL;
    size_t	len	= strlen(header->content_type->boundary);
    size_t	slen	= len + 2;
    size_t	elen	= len + 2 + 2;
    HeaderField	field	= MHF_START;

    boundary = (String)calloc(slen + 1, sizeof(char));
    if (boundary == NULL) {
      return;
    }
    bend = (String)calloc(elen + 1, sizeof(char));
    if (bend == NULL) {
      free((String)boundary); boundary = NULL;
      return;
    }
#if	HAVE_SNPRINTF
    snprintf(boundary, slen + 1, "--%s", header->content_type->boundary);
    snprintf(bend, elen + 1, "--%s--", header->content_type->boundary);
#else
    sprintf(boundary, "--%s", header->content_type->boundary);
    sprintf(bend, "--%s--", header->content_type->boundary);
#endif	/* HAVE_SNPRINTF */

    /* parse multipart actually */
    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf), mail->fin) != NULL) {
      /* end of multipart */
      if (strncmp(bend, buf, elen) == 0) {
	if (mail->option != OPT_PRINT && multipart != NULL) {
	  (void)fseek(multipart->ftmp, 0L, SEEK_SET);
	}
	break;	/* exit while() loop */
      }
      /* new part */
      if (strncmp(boundary, buf, slen) == 0) {
	if (mail->option != OPT_PRINT && multipart != NULL) {
	  (void)fseek(multipart->ftmp, 0L, SEEK_SET);
	}
	multipart		= newMultiPart(multipart, mail);
	multipart->filename	= getFileName(mail);
	if (mail->option != OPT_PRINT) {
	  int	fd = mkstemp(multipart->tmpfile);
	  if (fd == -1) {
	    error();
	  }
	  multipart->ftmp = fdopen(fd, "w+");
	  if (multipart->ftmp == NULL) {
	    error();
	  }
	  (void)unlink(multipart->tmpfile);
	}
	field	     = MHF_START;
	content_type = content_disposition = NULL;
	continue;
      }
      /* junk body in multipart body */
      if (multipart == NULL) {
	if (mail->option != OPT_PRINT) {
	  fputs(buf, body->ftmp);
	}
	continue;
      }

      /* multipart body */
      if (field == MHF_END) {
	if (mail->option != OPT_PRINT) {
	  fputs(buf, multipart->ftmp);
	}
	continue;
      }

      if (buf[0] == LF || (buf[0] == CR && buf[1] == LF)) {
	multipart->content_type = getContentType(content_type);
	multipart->content_disposition =
		getContentDisposition(content_disposition);
	field = MHF_END;	/* end of header */
	continue;
      }

      /* continued header */
      if (buf[0] == SPC || buf[0] == TAB) {
	for (value = buf; *value == SPC || *value == TAB; value++) {
	  ;	/* remove white spaces */
	}
	switch (field) {
	  case MHF_CONTENT_TYPE:
	    content_type = append(chomp(value), content_type);
	    continue;
	  case MHF_CONTENT_TRANSFER_ENCODING:
	    multipart->content_transfer_encoding =
		  append(chomp(value), multipart->content_transfer_encoding);
	    continue;
	  case MHF_CONTENT_DISPOSITION:
	    content_disposition = append(chomp(value), content_disposition);
	    continue;
	  case MHF_CONTENT_DESCRIPTION:
	    multipart->content_description =
		  append(chomp(value), multipart->content_description);
	    continue;
	  default:
	}
      }

      value = strchr(buf, ':');
      if (value != NULL) {
	for (++value; *value == SPC || *value == TAB; value++) {
	  ;	/* remove white spaces */
	}
      }

      chomp(value);
      if (strncasecmp(buf, MH_CONTENT_TYPE, length(MH_CONTENT_TYPE)) == 0) {
	content_type	= strdup(value);
	field		= MHF_CONTENT_TYPE;
      }
      else if (strncasecmp(buf, MH_CONTENT_TRANSFER_ENCODING,
			   length(MH_CONTENT_TRANSFER_ENCODING)) == 0) {
	multipart->content_transfer_encoding	= strdup(value);
	field					= MHF_CONTENT_TRANSFER_ENCODING;
      }
      else if (strncasecmp(buf, MH_CONTENT_DISPOSITION,
			   length(MH_CONTENT_DISPOSITION)) == 0) {
	content_disposition	= strdup(value);
	field			= MHF_CONTENT_DISPOSITION;
      }
      else if (strncasecmp(buf, MH_CONTENT_DESCRIPTION,
			   length(MH_CONTENT_DESCRIPTION)) == 0) {
	multipart->content_description	= strdup(value);
	field				= MHF_CONTENT_DESCRIPTION;
      }
      else {
	field = MHF_OTHERS;
      }
    } /* end of while() */

    free((String)boundary); boundary = NULL;
    free((String)bend); bend = NULL;
  }
  /* nonmultipart body */
  else {
    if (mail->option != OPT_PRINT) {
      while (fgets(buf, sizeof(buf), mail->fin) != NULL) {
	fputs(buf, body->ftmp);
      }
    }
  }
  if (mail->option != OPT_PRINT) {
    (void)fseek(body->ftmp, 0L, SEEK_SET);
  }
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	getFileName
 * Description:		return an new filename for a multipart body
 * Arguments:		Mail pointer
 * Return Value:	new filename
 */
static String
getFileName(mail)
	register Mail *	mail;
{
  String	fname	= NULL;
  size_t	len	= strlen(mail->filename) + 1 + 3 + 1;	/* .XXX */

  fname = (String)calloc(len, sizeof(char));
  if (fname == NULL) {
    error();
  }

#if	HAVE_SNPRINTF
  (void)snprintf(fname, len, "%s.%03d", mail->filename, mail->parts++);
#else
  (void)sprintf(fname, "%s.%03d", mail->filename, mail->parts++);
#endif	/* HAVE_SNPRINTF */

  return fname;
}

/*
 * Function Name:	newMultiPart
 * Description:		allocate new MultiPart memory
 * Arguments:		current MultiPart pointer, Mail pointer
 * Return Value:	new MultiPart pointer
 */
static MultiPart *
newMultiPart(curp, mail)
	MultiPart *	curp;
	Mail *		mail;
{
  MultiPart *	new = NULL;

  new = (MultiPart *)calloc(1, sizeof(MultiPart));
  if (new == NULL) {
    error();
  }
  new->tmpfile = strdup(TEMPLATE_FILE);

  if (curp == NULL) {
    new->next = NULL;
    mail->multipart = new;
  }
  else {
    curp->next = new;
  }

  return new;
}

/*
 * Function Name:	newReceivedFiled
 * Description:		parse Received field
 * Arguments:		Received field string
 * Return Value:	Received pointer
 */
static Received *
newReceivedFiled(received, curp)
	String		received;
	Received *	curp;
{
  Received *	new		= NULL;
  String	recv_from	= NULL;	/* From-domain */
  String	recv_by		= NULL;	/* By-domain */
  String	recv_via	= NULL;	/* Optional information */
  String	recv_with	= NULL;	/* Optional information */
  String	recv_id		= NULL;	/* Optional information */
  String	recv_for	= NULL;	/* Optional information */
  String	recv_date	= NULL;

  new = (Received *)calloc(1, sizeof(Received));
  if (new == NULL) {
    return curp;
  }

  recv_date = strchr(received, ';');
  if (recv_date != NULL) {
    *recv_date++ = EOL;
    for (; *recv_date == SPC || *recv_date == TAB; recv_date++) {
      ;	/* remove white spaces */
    }
  }

  recv_for	= strstr(received, "for ");
  recv_id	= strstr(received, "id ");
  recv_with	= strstr(received, "with ");
  recv_via	= strstr(received, "via ");
  recv_by	= strstr(received, "by ");
  recv_from	= strstr(received, "from ");

  if (recv_for != NULL) {
    if (*(recv_for - 1) == SPC) {
      *(recv_for - 1) = EOL;
    }
    recv_for += 4;
  }

  if (recv_id != NULL) {
    *(recv_id - 1) = EOL;
    recv_id += 3;
  }

  if (recv_with != NULL) {
    *(recv_with - 1) = EOL;
    recv_with += 5;
  }

  if (recv_via != NULL) {
    *(recv_via - 1) = EOL;
    recv_via += 4;
  }

  if (recv_by != NULL) {
    *(recv_by - 1) = EOL;
    recv_by += 3;
  }

  if (recv_from != NULL) {
    recv_from += 5;
  }

  new->recv_from	= xstrdup(recv_from);
  new->recv_by		= xstrdup(recv_by);
  new->recv_via		= xstrdup(recv_via);
  new->recv_with	= xstrdup(recv_with);
  new->recv_id		= xstrdup(recv_id);
  new->recv_for		= xstrdup(recv_for);
  new->recv_date	= xstrdup(recv_date);

  if (curp == NULL) {
    new->next = NULL;
  }
  else {
    new->next = curp;
  }

  return new;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * Function Name:	getContentType
 * Description:		parse Content-Type field
 * Arguments:		Content-Type
 * Return Value:	ContentType pointer
 */
static ContentType *
getContentType(content_type)
	String	content_type;
{
#define	PARAM_CHARSET	"charset="
#define	PARAM_BOUNDARY	"boundary="
#define	PARAM_NAME	"name="
  ContentType *	ctype	= NULL;
  String	param	= NULL;

  if (content_type == NULL) {
    return NULL;
  }

  ctype = (ContentType *)calloc(1, sizeof(ContentType));
  if (ctype == NULL) {
    error();
  }

  param = strchr(content_type, ';');
  if (param != NULL) {
    *param++ = EOL;
    for (; *param == SPC || *param == TAB ; param++) {
      ;	/* remove white spaces */
    }

    while ((param = commaSplit(param)) != NULL) {
      if (strncasecmp(param, PARAM_CHARSET, length(PARAM_CHARSET)) == 0) {
	int	offset = 0;
	if (param[length(PARAM_CHARSET)] == '"') {
	  param[strlen(param) - 1] = EOL;
	  offset = 1;
	}
	ctype->charset = strdup(param + length(PARAM_CHARSET) + offset);
      }
      else if (strncasecmp(param, PARAM_BOUNDARY,
			   length(PARAM_BOUNDARY)) == 0) {
	int	offset = 0;
	if (param[length(PARAM_BOUNDARY)] == '"') {
	  param[strlen(param) - 1] = EOL;
	  offset = 1;
	}
	ctype->boundary = strdup(param + length(PARAM_BOUNDARY) + offset);
      }
      else if (strncasecmp(param, PARAM_NAME, length(PARAM_NAME)) == 0) {
	int	offset = 0;
	if (param[length(PARAM_NAME)] == '"') {
	  param[strlen(param) - 1] = EOL;
	  offset = 1;
	}
	ctype->name = strdup(param + length(PARAM_NAME) + offset);
      }
      else {
	ctype->others = strdup(param);
      }

      param = NULL;	/* DON'T REMOVE THIS LINE. */
    }	/* end of while() */
  }
  ctype->media_type = strdup(content_type);

  return ctype;
}

/*
 * Function Name:	getContentDisposition
 * Description:		parse Content-Disposition field
 * Arguments:		Content-Disposition
 * Return Value:	ContentDisposition pointer
 */
static ContentDisposition *
getContentDisposition(content_disposition)
	register String	content_disposition;
{
#define	PARAM_FILENAME		"filename="
#define	PARAM_CREATION_DATE	"creation-date="
#define	PARAM_MODIFICATION_DATE	"modification-date="
#define	PARAM_READ_DATE		"read-date="
#define	PARAM_SIZE		"size="
  ContentDisposition *	cdsp	= NULL;
  String		param	= NULL;

  if (content_disposition == NULL) {
    return NULL;
  }

  cdsp = (ContentDisposition *)calloc(1, sizeof(ContentDisposition));
  if (cdsp == NULL) {
    error();
  }

  param = strchr(content_disposition, ';');
  if (param != NULL) {
    *param++ = EOL;
    for (; *param == SPC || *param == TAB ; param++) {
      ;	/* remove white spaces */
    }
    if (strncasecmp(param, PARAM_FILENAME, length(PARAM_FILENAME)) == 0) {
      int	offset = 0;
      if (param[length(PARAM_FILENAME)] == '"') {
	param[strlen(param) - 1] = EOL;
	offset = 1;
      }
      cdsp->filename = strdup(param + length(PARAM_FILENAME) + offset);
    }
    else if (strncasecmp(param, PARAM_CREATION_DATE,
			 length(PARAM_CREATION_DATE)) == 0) {
      int	offset = 0;
      if (param[length(PARAM_CREATION_DATE)] == '"') {
	param[strlen(param) - 1] = EOL;
	offset = 1;
      }
      cdsp->creation_date =
	strdup(param + length(PARAM_CREATION_DATE) + offset);
    }
    else if (strncasecmp(param, PARAM_MODIFICATION_DATE,
			 length(PARAM_MODIFICATION_DATE)) == 0) {
      int	offset = 0;
      if (param[length(PARAM_MODIFICATION_DATE)] == '"') {
	param[strlen(param) - 1] = EOL;
	offset = 1;
      }
      cdsp->modification_date =
	strdup(param + length(PARAM_MODIFICATION_DATE) + offset);
    }
    else if (strncasecmp(param, PARAM_READ_DATE,
			 length(PARAM_READ_DATE)) == 0) {
      int	offset = 0;
      if (param[length(PARAM_READ_DATE)] == '"') {
	param[strlen(param) - 1] = EOL;
	offset = 1;
      }
      cdsp->read_date = strdup(param + length(PARAM_READ_DATE) + offset);
    }
    else if (strncasecmp(param, PARAM_SIZE, length(PARAM_SIZE)) == 0) {
      int	offset = 0;
      if (param[length(PARAM_SIZE)] == '"') {
	param[strlen(param) - 1] = EOL;
	offset = 1;
      }
      cdsp->size = atoi(param + length(PARAM_SIZE) + offset);
    }
    else {
      cdsp->others = strdup(param);
    }
  }
  cdsp->disposition = strdup(content_disposition);

  return cdsp;
}


/*
 * Function Name:	commaSplit
 * Description:		split text with comma
 * Arguments:		splited text
 * Return Value:	new token
 */
static String
commaSplit(text)
	String	text;
{
  static String	src = NULL;
  String	ret = NULL;
  String	p   = NULL;

  if (text != NULL) {
    if (src != NULL) {
      free((String)src);
    }
    src = strdup(text);
  }
  if (src == NULL) {
    return NULL;
  }

  ret	= src;
  p	= strchr(src, ';');
check:
  if (p == NULL) {
    src = p;
    return ret;
  }

  if (*(p + 1) != SPC && *(p + 1) != TAB) {
    p = strchr(++p, ';');
    goto check;
  }
  *p++	= EOL;
  for (; *p == SPC || *p == TAB; p++) {
    ;	/* remove white spaces */
  }
  src	= p;
  return ret;
}

